# This file is part of openreviewio-py.
#
# openreviewio-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# openreviewio-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with openreviewio-py.  If not, see <https://www.gnu.org/licenses/>

from datetime import datetime, timezone
from pathlib import Path
import toml


class Status:

    def __init__(self, value, author, date: str = datetime.now(timezone.utc).isoformat()):
        """Status of a review.

        :param value: Value of the status. Accepted statuses are defined by the ORIO version. TODO: Insert here a dynamic list of available ones
        :param author: Author of the note.
        :param date: Note's time of creation, provided as ISO UTC (datetime.now(timezone.utc).isoformat()).
        """
        # Parameters
        status_allowed = toml.load(Path(__file__).parent.joinpath('orio_classes', 'Status.toml')).get('available')

        if value in status_allowed:
            self.state = value
        else:
            raise ValueError(f"Status '{value}' is not defined in the current version of OpenReviewIO standard."
                             f"Available statuses are: {', '.join(status_allowed)}.")

        self.author = author
        self.date = date

    # TODO Override print for status to print: Status ... written by ... at ...



# This file is part of openreviewio-py.
#
# openreviewio-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# openreviewio-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with openreviewio-py.  If not, see <https://www.gnu.org/licenses/>

from pathlib import Path
import toml


class Content:
    """Meta class for dynamically generated content objects from OpenReviewIO standard definition."""


def constructor(self, **kwargs):
    """Function to be used as __init__() for dynamically generated objects.
    If all required arguments hasn't been filled, it raises a TypeError.
    """

    given_args = set(dict(**kwargs).keys())

    # Get dynamically created attributes
    all_attributes = set([a for a in dir(self) if not a.startswith("_")])
    required_attributes = set([a for a in all_attributes if getattr(self, a) is None])

    # Test arguments, if not all required provided, raise the ones missing
    missing_args = required_attributes.difference(given_args)
    if required_attributes.difference(given_args):
        raise TypeError(
            f"All required arguments are not implemented for {self.__name__}, "
            f"please fill: '{', '.join(missing_args)}'.")

    # Set arguments, if doesn't exist, raise unknown arg
    for arg in given_args:
        if arg in all_attributes:
            setattr(self, arg, dict(**kwargs).get(arg))
        else:
            raise TypeError(
                f"Unknown argument for {self.__name__}: {arg}")

    # Specific cases
    if self.__name__ == 'Image':
        # Check mime for file extension
        file = Path(self.path_to_image)
        if file.suffix.replace('.', '') not in self.__mime__:
            raise TypeError(f"Image file '{file.suffix}' extension format is not valid. "
                            f"Please select either: {', '.join(sorted(self.__mime__))}.")


# Get all contents
all_contents = []
for file in Path(__file__).parent.joinpath('orio_classes').iterdir():
    # Load definition from TOML
    definition = toml.load(file)
    content = (file.stem, definition)

    # Sort them by dependencies, every inherited content follows its parent
    names = [d[0] for d in all_contents]
    if definition.get('type') in names:
        parent_index = names.index(definition['type'])
        all_contents.insert(parent_index + 1, content)
    else:
        all_contents.insert(0, content)

# Build classes from definitions
classes = []
for d in all_contents:
    # Get attributes
    if d[1].get('type') and d[1].get('type') not in ['Abstract', 'Object']:
        names = [d[0] for d in all_contents]
        parent_index = names.index(d[1].get('type'))
        parent_class = classes[parent_index]
    else:
        parent_class = object

    # Set definition parameters
    definition_parameters = {}
    for k, v in d[1].get('parameters', {}).items():
        if type(v) is dict:  # Process specificity
            if k == 'optional':
                for key, val in v.items():
                    definition_parameters[key] = ''
        else:
            definition_parameters[k] = None

    # Set definition attributes
    definition_attributes = {}
    for k, v in d[1].items():
        if type(v) is not dict:
            definition_attributes[f"__{k}__"] = v

    class_attributes = {**{'__init__': constructor, '__name__': d[0]}, **definition_attributes, **definition_parameters}

    # Generate class and add it as Content class attribute
    new_class = type(d[0], (parent_class,), class_attributes)
    setattr(Content, d[0], new_class)

    classes.append(new_class)


# This file is part of openreviewio-py.
#
# openreviewio-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# openreviewio-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with openreviewio-py.  If not, see <https://www.gnu.org/licenses/>


from __future__ import annotations
from typing import List, Union
from datetime import datetime, timezone
from .Content import Content


class Note:

    def __init__(self, author: str, date: str = datetime.now(timezone.utc).isoformat(),
                 contents: Union[Content, List[Content]] = None, parent: Note = None):
        """Contains contents.

        :param author: Author of the note.
        :param date: Note's time of creation, provided as ISO UTC (datetime.now(timezone.utc).isoformat()).
        :param contents: All note's contents. You can pass a single content at construction.
        :param parent: Note to reply to.
        """
        self.author = author
        self.date = date

        self.contents = []
        if contents:
            self.add_content(contents)

        self.parent = parent

    def add_content(self, content: Union[Content, List[Content]]):
        """Add content to the note.
        If the content is an Abstract one, a TypeError is raised.

        :param content: Content to add to the note.
        :return: Created content."""

        # Sentinel for checking if content type is valid
        if getattr(content, '__type__', '') == "Abstract":
            raise TypeError(f"Cannot create content from {content.__name__}, it is an abstract type.")

        # Add content, if list add all of them by running this function for each element to have the Abstract check
        if type(content) is list:
            for a in content:
                self.add_content(a)
        else:
            self.contents.append(content)



# This file is part of openreviewio-py.
#
# openreviewio-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# openreviewio-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with openreviewio-py.  If not, see <https://www.gnu.org/licenses/>

from typing import Union
from pathlib import Path
from .Content import Content
from .Note import Note
from .Status import Status
from .MediaReview import MediaReview


def load_media_review(media_review_path: Union[str, Path]) -> MediaReview:
    """Build a MediaReview object from the written media review at the given path.

    :param media_review_path: Path of written media review to load.
    :return: MediaReview object.
    """
    media_review_path = Path(media_review_path)

    if media_review_path.is_dir():
        review_folder_root = media_review_path
    else:
        review_folder_root = media_review_path.parent

    review_file = review_folder_root.joinpath('review.orio')

    # Build Review
    import xml.etree.ElementTree as ET

    review_xml = ET.parse(review_file).getroot()

    review = MediaReview(review_xml.attrib.get('media_path'))

    #   Statuses
    # Build statuses list
    statuses = []
    for status in review_xml.find('statuses').iter('status'):
        status_attributes = status.attrib
        statuses.append(Status(date=status_attributes.get('date'), author=status_attributes.get('author'),
                               value=status.text))

    current_status = statuses[-1]

    # Set statuses to review
    review.set_status_history(statuses)
    review.status = current_status

    # Notes
    for n in review_xml.find('notes').iter('note'):
        note_attributes = n.attrib
        note = Note(author=note_attributes.get('author'), date=note_attributes.get('date'),
                    parent=note_attributes.get('parent', ''))

        # Contents
        contents = []
        for a in n.find('contents').iter():
            if a.attrib:
                content = getattr(Content, a.tag)
                for k, v in a.attrib.items():
                    setattr(content, k, v)
                contents.append(content)
            note.contents = contents

            review.add_note(note)

    return review



# This file is part of openreviewio-py.
#
# openreviewio-py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# openreviewio-py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with openreviewio-py.  If not, see <https://www.gnu.org/licenses/>


from __future__ import annotations
from pathlib import Path
import shutil
import operator
from typing import List, Union
from .Status import Status
from .Note import Note
from .Content import Content
import xml.etree.ElementTree as ET


class MediaReview:

    media: Path
    _status_history: List[Status] = []  # Read only

    def __init__(self, media: Union[str, Path], status: Status = None, notes: List[Note] = None, ):
        """Review of a media.

        :param media: Path to the media the review is linked to.
        :param status: Review status.
        :param notes: Review's notes ordered by creation date. TODO Read only ?
        """
        self.media = Path(media)
        self.status = status
        self._status_history.append(status)

        self.notes = []
        if notes:
            self.add_note(notes)

    def add_note(self, note: Union[Note, List[Note]] = None):
        """Append a note (or all notes if a list is provided) it to the review's notes list.

        :param note: Note(s) to add."""

        if type(note) is list:
            for n in note:
                self.add_note(n)
        else:
            self.notes.append(note)

        # Sort notes by date
        self.notes = sorted(self.notes, key=operator.attrgetter('date'))

    def write(self, path: Union[str, Path] = '', include_attached_files: bool = True) -> Path:
        """Write the review in a folder according to the folder structure.
        If no path provided, the note is written next to the related media file in a folder 'media_name.ext.orio'.

        :param path: Folder path to write the review.
        :param include_attached_files: Copies all files related to contents in the created folder.
        :return: Path to the folder of the exported media review.
        """
        # Create review
        from inspect import getmembers

        # XML
        # create the file structure
        review = ET.Element('review', {'media_path': str(self.media), 'media_signature': 'd41d8cd98f00b204e9800998ecf8427e'})

        # Statuses
        statuses = ET.SubElement(review, 'statuses')
        for s in self.get_status_history():
            status = ET.Element('status', {'date': s.date, 'author': s.author})
            status.text = s.state
            statuses.append(status)

        # Notes
        notes = ET.SubElement(review, 'notes')
        files_to_copy = []
        for n in self.notes:
            note = ET.Element('note', {'date': n.date, 'author': n.author, 'parent': n.parent or ''})

            # Contents
            contents = ET.SubElement(note, 'contents')
            for a in n.contents:
                # Dynamically set parameters from members
                content = ET.SubElement(contents, a.__class__.__name__)
                for member in getmembers(a):
                    if not member[0].startswith('__'):
                        if type(member[1]) is Content.Image:
                            content.set('path_to_image', member[1].path_to_image)
                            # Keep files to copy
                            if include_attached_files:
                                files_to_copy.append(Path(member[1].path_to_image))
                        else:
                            content.set(member[0], str(member[1]))
            ET.SubElement(note, 'metadata')
            notes.append(note)

        #   Write to folder
        folder = Path(path).joinpath(f"{self.media.name}.orio")
        if not folder.is_dir():
            folder.mkdir()

        # Write to file
        file = folder.joinpath('review.orio')

        # create a new XML file with the results
        from xml.dom import minidom
        xmlstr = minidom.parseString(ET.tostring(review, encoding="utf-8")).toprettyxml(indent="   ")
        with file.open("w") as f:
            f.write(xmlstr)

        # Copy files if there have to be copied
        for f in files_to_copy:
            files_dir = folder.joinpath('files')
            if not files_dir.is_dir():
                files_dir.mkdir()

            shutil.copy2(f, files_dir.joinpath(f.name))

    def compare(self, review: MediaReview) -> MediaReview:
        """Compare the current review with the given one.
        TODO: Not implemented yet!

        :returns: A differential review with notes included in the current review that they aren't in the compared one.
        """
        pass

    def get_status_history(self) -> List[Status]:
        """Give the statuses history of the review.

        :returns: A list of statuses
        """
        return self._status_history

    def set_status_history(self, statuses_list):
        """Set the statuses history of the review.
        """
        self._status_history = statuses_list
